@@ -7,15 +7,27 @@ use std::marker::PhantomData;
 use std::os::unix::process::CommandExt;
 use std::process::{Child, Command};
 
-use nix::sys::{
-    ptrace,
-    wait::{self, WaitPidFlag, WaitStatus},
+use nix::{
+    errno::Errno,
+    sys::{
+        ptrace,
+        wait::{self, WaitPidFlag, WaitStatus},
+    },
 };
 
 use crate::error::{Error, Result, ResultExt};
 
+#[cfg(target_arch = "aarch64")]
+use crate::aarch64;
+
 #[cfg(target_arch = "x86_64")]
-use crate::x86::DebugRegister;
+use crate::x86;
+
+#[cfg(target_arch = "x86_64")]
+use x86::DebugRegister;
+
+#[cfg(target_arch = "aarch64")]
+pub type DebugRegisters = aarch64::user_hwdebug_state;
 
 pub use nix::unistd::Pid;
 pub use nix::sys::ptrace::Options;
@@ -23,6 +35,10 @@ pub use nix::sys::ptrace::Options;
 /// POSIX signal.
 pub use nix::sys::signal::Signal;
 
+/// Register state of a tracee.
+#[cfg(target_arch = "aarch64")]
+pub type Registers = aarch64::user_pt_regs;
+
 /// Register state of a tracee.
 #[cfg(target_arch = "x86_64")]
 pub type Registers = libc::user_regs_struct;
@@ -32,6 +48,10 @@ pub type Siginfo = libc::siginfo_t;
 
 const WALL: Option<WaitPidFlag> = Some(WaitPidFlag::__WALL);
 
+/// Linux constant defined in `include/uapi/linux/elf.h`.
+#[cfg(target_arch = "aarch64")]
+const NT_PRSTATUS: i32 = 0x1;
+
 /// A _ptrace-stop_, a tracee state in which it is stopped and ready to accept ptrace
 /// commands.
 ///
@@ -118,11 +138,45 @@ impl Tracee {
         Ok(ptrace::getregs(self.pid).died_if_esrch(self.pid)?)
     }
 
+    #[cfg(target_arch = "aarch64")]
+    pub fn registers(&self) -> Result<Registers> {
+
+        let mut data = std::mem::MaybeUninit::uninit();
+        let mut rv = libc::iovec {
+            iov_base: &mut data as *mut _ as *mut libc::c_void,
+            iov_len: std::mem::size_of::<Registers>(),
+        };
+
+        let res = unsafe {
+            libc::ptrace(libc::PTRACE_GETREGSET, self.pid, NT_PRSTATUS, &mut rv as *mut _ as *mut libc::c_void)
+        };
+
+        Errno::result(res)?;
+
+        Ok( unsafe { data.assume_init() } )
+    }
+
     #[cfg(target_arch = "x86_64")]
     pub fn set_registers(&mut self, regs: Registers) -> Result<()> {
         Ok(ptrace::setregs(self.pid, regs).died_if_esrch(self.pid)?)
     }
 
+    #[cfg(target_arch = "aarch64")]
+    pub fn set_registers(&mut self, regs: Registers) -> Result<()> {
+        let mut rv = libc::iovec {
+            iov_base: &regs as *const _ as *const libc::c_void as *mut libc::c_void,
+            iov_len: std::mem::size_of::<Registers>(),
+        };
+
+        let res = unsafe {
+            libc::ptrace(libc::PTRACE_SETREGSET, self.pid, NT_PRSTATUS, &mut rv as *mut _ as *mut libc::c_void)
+        };
+
+        Errno::result(res)?;
+
+        Ok(())
+    }
+
     pub fn read_memory(&mut self, addr: u64, len: usize) -> Result<Vec<u8>> {
         let mut data = Vec::with_capacity(len);
         data.resize(len, 0);
@@ -178,6 +232,22 @@ impl Tracee {
         }
     }
 
+    #[cfg(target_arch = "aarch64")]
+    pub fn debug_registers(&self, regtype: aarch64::DebugRegisterType) -> Result<DebugRegisters> {
+        let mut data = std::mem::MaybeUninit::uninit();
+        let mut rv = libc::iovec {
+            iov_base: &mut data as *mut _ as *mut libc::c_void,
+            iov_len: std::mem::size_of::<aarch64::user_hwdebug_state>(),
+        };
+        let res = unsafe {
+            libc::ptrace(libc::PTRACE_GETREGSET, self.pid, regtype, &mut rv as *mut _ as *mut libc::c_void)
+        };
+
+        Errno::result(res)?;
+
+        Ok(unsafe { data.assume_init() })
+    }
+
     #[cfg(target_arch = "x86_64")]
     pub fn set_debug_register(&self, dr: DebugRegister, data: u64) -> Result<()> {
         let index = 8 * u64::from(dr);
@@ -189,6 +259,21 @@ impl Tracee {
         }
     }
 
+    #[cfg(target_arch = "aarch64")]
+    pub fn set_debug_registers(&self, regtype: aarch64::DebugRegisterType, mut state: DebugRegisters) -> Result<()> {
+        let mut rv = libc::iovec {
+            iov_base: &mut state as *mut _ as *mut libc::c_void,
+            iov_len: std::mem::size_of::<aarch64::user_hwdebug_state>(),
+        };
+        let res = unsafe {
+            libc::ptrace(libc::PTRACE_SETREGSET, self.pid, regtype, &mut rv as *mut _ as *mut libc::c_void)
+        };
+
+        Errno::result(res)?;
+
+        Ok(())
+    }
+
     #[cfg(target_arch = "x86_64")]
     fn peek_user(&self, off: u64) -> Result<u64> {
         // SAFETY: `off` does not require validation, because it is not actually used as a
@@ -291,7 +376,7 @@ impl Ptracer {
     pub fn spawn(&mut self, mut cmd: Command) -> Result<Child> {
         // On fork, request `PTRACE_TRACEME`.
         unsafe {
-            cmd.pre_exec(|| ptrace::traceme().map_err(as_ioerror))
+            cmd.pre_exec(|| ptrace::traceme().map_err(|err| io::Error::from_raw_os_error(err as i32)))
         };
 
         let child = cmd.spawn()?;
@@ -327,7 +412,7 @@ impl Ptracer {
         let status = match wait::waitpid(None, WALL) {
             Ok(status) =>
                 status,
-            Err(nix::Error::Sys(errno)) if errno == nix::errno::Errno::ECHILD =>
+            Err(errno) if errno == Errno::ECHILD =>
                 // No more children to wait on: we're done.
                 return Ok(None),
             Err(err) =>
@@ -650,8 +735,6 @@ fn is_group_stop(pid: Pid, sig: Signal) -> Result<bool> {
 
     match sig {
         SIGSTOP | SIGTSTP | SIGTTIN | SIGTTOU => {
-            use nix::{errno::Errno, Error};
-
             // Possible group-stop. Check `siginfo` to disambiguate.
             //
             // From the manual:
@@ -661,7 +744,7 @@ fn is_group_stop(pid: Pid, sig: Signal) -> Result<bool> {
             //     ("no such process") if a SIGKILL killed the tracee.)
             //
             match ptrace::getsiginfo(pid) {
-                Err(Error::Sys(Errno::EINVAL)) =>
+                Err(Errno::EINVAL) =>
                     Ok(true),
                 Err(err) =>
                     Err(err.into()),
@@ -682,14 +765,3 @@ fn is_group_stop(pid: Pid, sig: Signal) -> Result<bool> {
         },
     }
 }
-
-// Only intended for the result of `ptrace::traceme()`.
-fn as_ioerror(err: nix::Error) -> io::Error {
-    if let Some(errno) = err.as_errno() {
-        io::Error::from_raw_os_error(errno as i32)
-    } else {
-        // Should be unreachable when used with `ptrace::traceme()`, since its `Result`
-        // comes from an internal call to `nix::Errno::result()`.
-        io::Error::new(io::ErrorKind::Other, err)
-    }
-}
